#include <Windows.h>
#include <iostream>
#include "SpaceMousePro.h"

// Window message callback.
LRESULT CALLBACK EventHandler(HWND, unsigned, WPARAM, LPARAM);
int GetHIDDeviceList();

int main(int argc, char** argv) {
	HINSTANCE instance = GetModuleHandle(0);

	std::ios_base::sync_with_stdio(false); 	// Make std::cout faster:
	std::cout << "INIT SPACE MOUSE" << std::endl;

	// Create empty window as message sink from hardware
	LPCWSTR class_name = L"MessageSink"; // Create message-only window:
	WNDCLASS window_class = {};
	window_class.lpfnWndProc = EventHandler;
	window_class.hInstance = instance;
	window_class.lpszClassName = class_name;

	if (!RegisterClass(&window_class)) {
		std::cerr << "RegisterClass ERROR" << std::endl;
		return -1;
	}

	HWND window = CreateWindow(class_name, L"MessageSink", 0, 0, 0, 0, 0, HWND_MESSAGE, 0, 0, 0);
	if (window == nullptr) {
		std::cerr << "Create message sink Window ERROR" << std::endl;
		return -1;
	}

#ifdef SHOW_ALL_INPUT_DEVICE
	GetHIDDeviceList();
#endif // SHOW_ALL_INPUT_DEVICE

	// Register input device[s]
	RAWINPUTDEVICE rid[1];
	rid[0].usUsagePage = 0x01;
	rid[0].usUsage = 0x08;
	rid[0].dwFlags = RIDEV_INPUTSINK;
	rid[0].hwndTarget = window;
	RegisterRawInputDevices(rid, 1, sizeof(rid[0]));

	// Main loop:
	MSG event;
	bool quit = false;

	while (!quit) {
		while (PeekMessage(&event, 0, 0, 0, PM_REMOVE)) {
			if (event.message == WM_QUIT) {
				quit = true;
				break;
			}
			TranslateMessage(&event);
			DispatchMessage(&event);
		}
	}

	return 0;
}

LRESULT CALLBACK EventHandler(HWND hwnd, unsigned event, WPARAM wparam, LPARAM lparam) {
	switch (event) {
		case WM_DESTROY:
			PostQuitMessage(0);
			return 0;
		case WM_INPUT: {
			unsigned size = sizeof(RAWINPUT);
			static RAWINPUT raw[sizeof(RAWINPUT)];

			RID_DEVICE_INFO rdiDeviceInfo;
			rdiDeviceInfo.cbSize = sizeof(RID_DEVICE_INFO);
			UINT nBufferSize = rdiDeviceInfo.cbSize;

			GetRawInputData((HRAWINPUT)lparam, RID_INPUT, raw, &size, sizeof(RAWINPUTHEADER));

			if (GetRawInputDeviceInfo(raw->header.hDevice, RIDI_DEVICEINFO, &rdiDeviceInfo, &nBufferSize) < 0) {
				std::cerr << "Unable to read Device Info" << std::endl;
				return 0;
			}

			// Test if is a spacemouse
			if (raw->header.dwType == RIM_TYPEHID) {
				if (rdiDeviceInfo.hid.dwVendorId == SPACE_MOUSE_VENDOR_ID && rdiDeviceInfo.hid.dwProductId == SPACE_MOUSE_PRODUCT_ID) {
					processSpaceMouseProRawData(raw->data.hid.bRawData, raw->data.hid.dwSizeHid);
				}
			}
		} return 0;
	}

	return DefWindowProc(hwnd, event, wparam, lparam); // Run default message processor for any missed events:
}

int GetHIDDeviceList() {
	UINT nDevices = 0;
	int nResult = 0;
	PRAWINPUTDEVICELIST pRawInputDeviceList;

	GetRawInputDeviceList(NULL, &nDevices, sizeof(RAWINPUTDEVICELIST)); // Get Number Of Devices

	if (nDevices < 1) { return 0; }

	pRawInputDeviceList = new RAWINPUTDEVICELIST[sizeof(RAWINPUTDEVICELIST) * nDevices]; // Allocate Memory For Device List
	if (pRawInputDeviceList == NULL) { return -1; }

	nResult = GetRawInputDeviceList(pRawInputDeviceList, &nDevices, sizeof(RAWINPUTDEVICELIST)); // Fill Device List Buffer
	if (nResult < 0) {
		std::cerr << "Error get Device list" << std::endl;
		delete[] pRawInputDeviceList; // Clean Up
		return 0;
	}

	for (UINT i = 0; i < nDevices; i++) {
		UINT nBufferSize = 0; // Get Character Count For Device Name
		nResult = GetRawInputDeviceInfo(pRawInputDeviceList[i].hDevice, // Device
			RIDI_DEVICENAME,                // Get Device Name
			NULL,                           // NO Buff, Want Count!
			&nBufferSize);                 // Char Count Here!
		if (nResult < 0) { // Error get Device Name character count
			std::cerr << "Error get Device Name character count" << std::endl;
			continue; // Goto next device
		}

		// Get Name
		UINT nameSize = nBufferSize + 1;
		WCHAR* wcDeviceName = new WCHAR[nameSize]; // Allocate Memory For Device Name
		if (wcDeviceName == NULL) { continue; }
		nResult = GetRawInputDeviceInfo(pRawInputDeviceList[i].hDevice, // Device
			RIDI_DEVICENAME,                // Get Device Name
			wcDeviceName,                   // Get Name!
			&nBufferSize);                 // Char Count
		if (nResult < 0) { // Error get Device Name
			std::cerr << "Error get Device Name" << std::endl;
			delete[] wcDeviceName; // Clean Up
			continue; // Goto next device
		}

		// Set Device Info & Buffer Size
		RID_DEVICE_INFO rdiDeviceInfo;
		rdiDeviceInfo.cbSize = sizeof(RID_DEVICE_INFO);
		nBufferSize = rdiDeviceInfo.cbSize;

		nResult = GetRawInputDeviceInfo(pRawInputDeviceList[i].hDevice, RIDI_DEVICEINFO, &rdiDeviceInfo, &nBufferSize); // Get Device Info
		if (nResult < 0) {
			std::cerr << "Unable to read Device Info" << std::endl;
			continue; // Goto next device
		}

		if (rdiDeviceInfo.dwType == RIM_TYPEMOUSE) {
			std::cout << std::endl << "------ Found Mouse ------" << std::endl;
			std::wcout << L"Device Name: " << wcDeviceName << std::endl;
			std::cout << "Mouse ID: " << rdiDeviceInfo.mouse.dwId << std::endl;
			std::cout << "Mouse buttons: " << rdiDeviceInfo.mouse.dwNumberOfButtons << std::endl;
			std::cout << "Mouse sample rate (Data Points): " << rdiDeviceInfo.mouse.dwSampleRate << std::endl;
			if (rdiDeviceInfo.mouse.fHasHorizontalWheel) {
				std::cout << "Mouse has horizontal wheel" << std::endl;
			}
			else {
				std::cout << "Mouse does not have horizontal wheel" << std::endl;
			}
		}
		else if (rdiDeviceInfo.dwType == RIM_TYPEKEYBOARD) {
			std::cout << std::endl << "------ Found keyboard ------" << std::endl;
			std::wcout << L"Device Name: " << wcDeviceName << std::endl;
			std::cout << "Keyboard mode: " << rdiDeviceInfo.keyboard.dwKeyboardMode << std::endl;
			std::cout << "Number of function keys: " << rdiDeviceInfo.keyboard.dwNumberOfFunctionKeys << std::endl;
			std::cout << "Number of indicators: " << rdiDeviceInfo.keyboard.dwNumberOfIndicators << std::endl;
			std::cout << "Number of keys total: " << rdiDeviceInfo.keyboard.dwNumberOfKeysTotal << std::endl;
			std::cout << "Type of the keyboard: " << rdiDeviceInfo.keyboard.dwType << std::endl;
			std::cout << "Subtype of the keyboard: " << rdiDeviceInfo.keyboard.dwSubType << std::endl;
		}
		else if (rdiDeviceInfo.dwType == RIM_TYPEHID) {
			std::cout << std::endl << "------ Found HID Input ------" << std::endl;
			std::wcout << L"Device Name: " << wcDeviceName << std::endl;
			std::cout << "Vendor Id:" << rdiDeviceInfo.hid.dwVendorId << std::endl;
			std::cout << "Product Id:" << rdiDeviceInfo.hid.dwProductId << std::endl;
			std::cout << "Version No:" << rdiDeviceInfo.hid.dwVersionNumber << std::endl;
			std::cout << "Usage for the device: " << rdiDeviceInfo.hid.usUsage << std::endl;
			std::cout << "Usage Page for the device: " << rdiDeviceInfo.hid.usUsagePage << std::endl;
		}

		delete[] wcDeviceName;
	}

	delete[] pRawInputDeviceList;
	return nDevices;
}